#ifndef BASE_H
#define BASE_H

extern "C"
{
#include <lip.h>
}
#include <fstream.h>
#include <stdlib.h>


// Types
///////////////////////////////////////////////////////////////////////////////

typedef unsigned short	ushort;
typedef unsigned long	ulong;		// XXX - unsigned int?

#ifdef sgi
typedef int bool;
#define true 1
#define false 0
#endif

// Component of a list of factors: the number itself, its state
// (prime, nonprime, unknown...), and its multiplicity.
class IntElem
{
friend class IntList;
public:
			IntElem() { mCount = 0; }
			IntElem(verylong n, ushort count, ushort state)
					{ Create(); zcopy(n, mNum);
						mCount = count;
						mState = state; }
			IntElem(ulong n, ushort count, ushort state)
					{ Create(); zuintoz(n, mNum);
						mCount = count;
						mState = state; }
			~IntElem();

	void		SetNum(verylong n) { zcopy(n, mNum); }
	void		SetNum(ulong n) { zuintoz(n, mNum); }
	void		SetCount(ushort count) { mCount = count; }
	void		SetState(ushort state) { mState = state; }

	const verylong*	GetNum() const { return mNum; }
	ushort		GetCount() const { return mCount; }
	ushort		GetState() const { return mState; }

private:
	void		Create();

	verylong*	mNum;
	int		mNumIndex;
	ushort		mCount;
	ushort		mState;
	IntElem*	mPrev;
	IntElem*	mNext;
};


typedef IntElem* POSITION;

extern ostream& operator << (ostream& o, const IntList& l);
extern ostream& operator << (ostream& o, verylong n);


class IntList
{
public:
			IntList();
			~IntList();

	bool		IsEmpty() const { return GetSize() == 0; }
	int		GetSize() const { return mSize; }

	POSITION	GetHeadPosition() const { return mHead; }
	IntElem*	GetNext(POSITION& pos) const;

	bool		Add(verylong n, ushort count, ushort state);
	bool		Add(ulong n, ushort count, ushort state);
	bool		Add(IntElem* elem);
	void		Remove();
	void		Clear();

	const verylong*	GetNum() const { return mTail->GetNum(); }
	ushort		GetCount() const { return mTail->GetCount(); }
	ushort		GetState() const { return mTail->GetState(); }

	void		Print() const { cout << *this; }
	void		Multiply(verylong* r) const;
	bool		Verify(verylong n) const;
	void		Merge(IntList& l1, IntList& l2);

protected:

	IntElem*	mHead;
	IntElem*	mTail;
	int		mSize;

};


typedef void (*checkmark_func_t)(ostream& checkfile);


// The list with prime number diffs up to MaxUpperBound
///////////////////////////////////////////////////////////////////////////////

const ulong MaxUpperBound = 10000000;
extern ulong* primeDiff;


// Constants that describe the state of a number or factor
///////////////////////////////////////////////////////////////////////////////

static const ushort	NumPrime	= 100;
static const ushort	NumComposite	= 101;
static const ushort	NumProbPrime	= 102;


// Variables
///////////////////////////////////////////////////////////////////////////////

extern int	trace_level;
extern bool	checkpoint;


// Functions
///////////////////////////////////////////////////////////////////////////////

#define zislong(n) (n == 0 || n[0] >= -1 && n[0] <= 1)
#define zeven(n) (zodd(n) == 0)
#define ztwo(n) zuintoz(2, n);
#define zthree(n) zuintoz(3, n);
#define zssub(n, dec, m) zsadd(n, -dec, m)
#define zsinc(n, inc) zsadd(n, inc, &n)
#define zinc(n, inc) zadd(n, inc, &n)
#define zsdec(n, dec) zsadd(n, -dec, &n)
#define zdec(n, dec) zsub(n, dec, &n)

extern double ratio(verylong x, verylong y);
extern void powersum(verylong p, ulong n, verylong* r);
extern void lucas(verylong m, verylong n, long p, long q,
	verylong* u, verylong* v);
extern void lucas2(ulong m, verylong n,
	verylong p, verylong* u, verylong* v);
extern ushort fermat_test(verylong n, verylong base);
extern ushort strong_test(verylong n, verylong base);
extern ushort num_status(verylong n, ulong u);
extern ushort factor_brute(verylong n, ulong l, ulong u, IntList& r);
extern verylong* max3(verylong* n1, verylong* n2, verylong* n3);
extern void factor_brute3(verylong n, ulong l, ulong u,
	IntList& r1, IntList& r2, IntList& r3);
extern void print_state(ushort state);
extern bool InitBase();
extern void factorsum(verylong n, ulong upper, verylong* r, IntList& l);
extern void e_sieve(verylong n, verylong count,
	verylong upper, verylong* r);
extern void print_sieve(verylong n, verylong sieve);
extern void fac(ulong n, verylong* r);
extern ulong gcd(ulong a, ulong b);
extern void binomial(ulong n, ulong k, verylong* r);

extern ushort theorem8(verylong n, const IntList& r,
	ushort state, ulong u, verylong* factor);
extern ushort theorem9(verylong n, const IntList& r,
	ushort state, ulong u, verylong* factor);
extern ushort theorem10(verylong n, const IntList& r,
	ushort state, ulong u, verylong* factor);
extern ushort theorem11(verylong n, const IntList& r,
	ushort state, ulong u, verylong* factor);
extern ushort prove_prime(verylong n, ulong l, ulong u, verylong* factor);

extern ushort factor(verylong n, ulong l, ulong u, IntList& r);
extern ushort mc_factor(verylong n, verylong lim, verylong* factor);
extern ushort pollard_factor(verylong n, ulong b1, ulong b2,
	verylong* factor);
extern ushort pplus1_factor(verylong n, ulong b1, ulong b2,
	verylong* factor);

extern void ts(verylong n);
extern void FriendSearch(verylong start, int onlyone);
extern void AugmentedSearch(verylong start, int onlyone);
extern void QuasiSearch(verylong start, int onlyone);
extern void Carmichael(verylong start, verylong base);

extern void InitCheckMark(char* progName);
extern void WriteCheckMark(checkmark_func_t cmf);

#endif /* BASE_H */

