#include <stdio.h>
#include <stdlib.h>
#include "base.h"

// Searches Carmichael numbers

static verylong carm_n = 0;

static void CheckCarmichael(ostream& checkfile)
{
	checkfile << carm_n;
}

static verylong base2 = 0;
static verylong base3 = 0;
static verylong base5 = 0;
static verylong base7 = 0;
static verylong base11 = 0;
static verylong base13 = 0;
static verylong base17 = 0;
static verylong base19 = 0;

static void CTest(verylong n)
{
	static verylong temp = 0;
	bool print = false;
	char t[8];
	int i;

	for (i = 0; i < 8; i++)
		t[i] = ' ';

	i = 0;
	zgcd(carm_n, base2, &temp);
	if (zscompare(temp, 1) > 0)
		t[i] = 'g';
	else if (strong_test(carm_n, base2) == NumProbPrime)
	{
		t[i] = 's';
		print = true;
	}
	else if (fermat_test(carm_n, base2) == NumProbPrime)
	{
		t[i] = 'w';
		print = true;
	}

	i++;
	zgcd(carm_n, base3, &temp);
	if (zscompare(temp, 1) > 0)
		t[i] = 'g';
	else if (strong_test(carm_n, base3) == NumProbPrime)
	{
		t[i] = 's';
		print = true;
	}
	else if (fermat_test(carm_n, base3) == NumProbPrime)
	{
		t[i] = 'w';
		print = true;
	}

	i++;
	zgcd(carm_n, base5, &temp);
	if (zscompare(temp, 1) > 0)
		t[i] = 'g';
	else if (strong_test(carm_n, base5) == NumProbPrime)
	{
		t[i] = 's';
		print = true;
	}
	else if (fermat_test(carm_n, base5) == NumProbPrime)
	{
		t[i] = 'w';
		print = true;
	}

	i++;
	zgcd(carm_n, base7, &temp);
	if (zscompare(temp, 1) > 0)
		t[i] = 'g';
	else if (strong_test(carm_n, base7) == NumProbPrime)
	{
		t[i] = 's';
		print = true;
	}
	else if (fermat_test(carm_n, base7) == NumProbPrime)
	{
		t[i] = 'w';
		print = true;
	}

	i++;
	zgcd(carm_n, base11, &temp);
	if (zscompare(temp, 1) > 0)
		t[i] = 'g';
	else if (strong_test(carm_n, base11) == NumProbPrime)
	{
		t[i] = 's';
		print = true;
	}
	else if (fermat_test(carm_n, base11) == NumProbPrime)
	{
		t[i] = 'w';
		print = true;
	}

	i++;
	zgcd(carm_n, base13, &temp);
	if (zscompare(temp, 1) > 0)
		t[i] = 'g';
	else if (strong_test(carm_n, base13) == NumProbPrime)
	{
		t[i] = 's';
		print = true;
	}
	else if (fermat_test(carm_n, base13) == NumProbPrime)
	{
		t[i] = 'w';
		print = true;
	}

	i++;
	zgcd(carm_n, base17, &temp);
	if (zscompare(temp, 1) > 0)
		t[i] = 'g';
	else if (strong_test(carm_n, base17) == NumProbPrime)
	{
		t[i] = 's';
		print = true;
	}
	else if (fermat_test(carm_n, base17) == NumProbPrime)
	{
		t[i] = 'w';
		print = true;
	}

	i++;
	zgcd(carm_n, base19, &temp);
	if (zscompare(temp, 1) > 0)
		t[i] = 'g';
	else if (strong_test(carm_n, base19) == NumProbPrime)
	{
		t[i] = 's';
		print = true;
	}
	else if (fermat_test(carm_n, base19) == NumProbPrime)
	{
		t[i] = 'w';
		print = true;
	}

	if (print)
	{
		cout << n << "\t";
		for (i = 0; i < 8; i++)
		{
			if (t[i] == ' ')
				cout << "\t";
			else
				cout << "\t" << t[i];
		}
		cout << endl;
	}

} // CTest

/*
 * Algorithm:
 * Search all primes in a block of numbers. For all non-primes do
 * strong_test and fermat_test with small bases.
 * XXX - It might be faster to test all pseudo primes for primality
 * rather than sieving.
 */
static void Carmichael2(verylong start)
{
	ulong block_len = 100000;
	static verylong sieve = 0;
	static verylong temp2 = 0;
	static verylong temp3 = 0;

	zcopy(start, &carm_n);

	zuintoz(2, &base2);
	zuintoz(3, &base3);
	zuintoz(5, &base5);
	zuintoz(7, &base7);
	zuintoz(11, &base11);
	zuintoz(13, &base13);
	zuintoz(17, &base17);
	zuintoz(19, &base19);

	if (zeven(carm_n))
                zsinc(carm_n, 1);

        // Limitation of the block length so that the precondition for
	// e_sieve is met.
	if (zscompare(carm_n, 400) < 0)
	{
		zssub(carm_n, 1, &temp2);
		zmul(carm_n, temp2, &temp3);
		zsadd(temp3, 2, &temp2);
		block_len = ztouint(temp2);
	}

	while (1)
	{
		ulong index = 0;
		ulong end = block_len / 2;

		if (trace_level > 3)
			cout << "Sieving..." << endl;

                zuintoz(block_len, &temp2);
		e_sieve(carm_n, temp2, carm_n, &sieve);

		if (trace_level > 1)
			cout << "Testing " << carm_n << endl;

		while (index < end)
		{
			if (zbit(sieve, index))
			{
				CTest(carm_n);
			}
                        zsinc(carm_n, 2);
			index++;
		}

		if (checkpoint)
			WriteCheckMark(CheckCarmichael);
	}

} // Carmichael2

int main(int argc, char** argv)
{
	verylong n = 0;

	if (argc < 2)
	{
		printf("Usage: %s start\n", argv[0]);
		return 1;
	}

	if (zsread(argv[1], &n) == 0 || zsign(n) <= 0)
	{
		printf("Usage: %s start\n", argv[0]);
		return 1;
	}

	trace_level = 1;

	InitBase();
	InitCheckMark(argv[0]);
	Carmichael2(n);

	return 0;

}

