#include <stdio.h>
#include <stdlib.h>
#include "base.h"

// Searches Carmichael Numbers

extern short incr_list[];
extern int incr_len;
const int incr_period = 2310;


static void Carmichael3(verylong n, verylong lower, verylong upper)
{
	static verylong	base = 0;
	static verylong	u = 0;
	static verylong	l = 0;
	static verylong	temp = 0;
	int		strong = 1;
	int		t = 0;

	ztwo(&base);

	if (upper > n)
		zcopy(n, &u);
	else
		zcopy(upper, &u);

	zcopy(lower, &l);
	if (zcompare(l, u) > 0)
	{
		cout << " Nothing to do" << endl;
		return;
	}

	if (zsign(l) > 0)
	{
		ulong m = zsmod(l, incr_period);
		zsdec(l, m);
		zsinc(l, 13);
		zcopy(l, &base);
		t = 5;
	}

	while (zcompare(base, u) < 0)
	{
		zgcd(n, base, &temp);
		if (zscompare(temp, 1) == 0)
		{
			if (trace_level > 4)
				cout << "Testing " << base << endl;

			if (strong)
			{
				if (strong_test(n, base) != NumProbPrime)
				{
					strong = 0;
					if (fermat_test(n, base) != NumProbPrime)
						break;		// No carmichael
				}
			}
			else
			{
				if (fermat_test(n, base) != NumProbPrime)
					break;
			}
		}

		if (t >= incr_len)
			t = 5;  	// Next cycle
		zsinc(base, incr_list[t++]);
							 
	}

	if (zcompare(base, u) >= 0)
	{
		if (zcompare(u, n) == 0)
			cout << ": " << (strong ? "Strong" : "Weak")
				<< " Carmichael number" << endl;
		else
			cout << ": Possible " << (strong ? "Strong" : "Weak")
				<< " Carmichael number" << endl;
	}
	else
		cout << endl;

} // Carmichael3

int Usage(char** argv, int errCode)
{
	printf("Usage: %s number [base-lower [base-upper]]\n", argv[0]);
	return errCode;
}

int main(int argc, char** argv)
{
	verylong n = 0;
	verylong lower = 0;
	verylong upper = 0;
	IntList r;

	if (argc < 2)
		return Usage(argv, 1);

	if (zsread(argv[1], &n) == 0 || zscompare(n, 1) <= 0)
		return Usage(argv, 1);

	if (argc > 2)
	{
		if (zsread(argv[2], &lower) == 0 || zsign(lower) < 0)
			return Usage(argv, 1);
	}

	zuintoz(200000, &upper);
	if (argc > 3)
	{
		if (zsread(argv[3], &upper) == 0
			|| zcompare(upper, lower) < 0)
			return Usage(argv, 1);
	}

	trace_level = 0;

	if (InitBase() == false)
		return 1;
	factor(n, 0, 1000000, r);
	if (r.Verify(n) == false)
		return 1;
	cout << n;
	if (zcompare(*r.GetNum(), n) == 0)
	{
		if (r.GetState() == NumPrime)
			cout << " Prime number" << endl;
		else
			cout << " Error: State of " << n << " Unknown" << endl;
	}
	else
	{
		cout << " =" << r;
		Carmichael3(n, lower, upper);
	}

	return 0;

}

