#ifdef sgi
#define _BSD_SIGNALS
#endif

#include <limits.h>
#include <setjmp.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include "base.h"

bool checkpoint;

static const int CHECK_INTERVAL = 60 * 10;		// 10 minutes

static char prog[PATH_MAX];
static char name[PATH_MAX];
static char oldname[PATH_MAX];
static ofstream checkfile;
static checkmark_func_t checkfunc = NULL;
static sigjmp_buf jmpbuf;
static volatile sig_atomic_t canjump = 0;

void WriteCheckMark(checkmark_func_t cmf)
{
	checkfunc = cmf;
	checkpoint = false;

	strcpy(name, prog);
	strcat(name, ".check");
	strcpy(oldname, name);
	strcat(oldname, ".old");

	rename(name, oldname);
	checkfile.open(name, ios::out | ios::trunc);
	checkfile << prog << " ";
	(*cmf)(checkfile);
	checkfile << endl;
	if (!checkfile)
		unlink(name);
	checkfile.close();
} // WriteCheckMark

#ifdef sgi
void AlarmHandler(...)
#else
void AlarmHandler(int sig)
#endif
{
	// We can't write directly into a file here because malloc and stdio
	// functions are not reentrant.
	checkpoint = true;

	signal(SIGALRM, AlarmHandler);
	alarm(CHECK_INTERVAL);
} // AlarmHandler

#ifdef sgi
void TermHandler(...)
#else
void TermHandler(int sig)
#endif
{
	if (canjump == 0)
		return;
	canjump = 0;
	siglongjmp(jmpbuf, 1);
} // TermHandler

void InitCheckMark(char* progName)
{
	char* p;

	if ((p = strrchr(progName, '/')) != NULL)
		strcpy(prog, &p[1]);
	else
		strcpy(prog, progName);

	checkpoint = false;
	signal(SIGALRM, AlarmHandler);
	signal(SIGINT, TermHandler);
	if (sigsetjmp(jmpbuf, 1))
	{
		if (checkfunc != NULL)
			WriteCheckMark(checkfunc);
		exit(0);
	}

	canjump = 1;		// siglongjmp is now allowed
	alarm(10);		// First alarm after 10 seconds

} // InitCheckMark
