#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "base.h"


// Generates a list with differences between primes from 2 to MaxUpperBound.
// This list is dumped to file primediff.dat.

extern short	incr_list[];
extern int	incr_len;

ulong*		p;
size_t		length = 0;

verylong	nOld = 0;


// n odd
void eval(verylong sieve, verylong* n, verylong count)
{
	static verylong cnt = 0;

	zcopy(count, &cnt);

	while (zsign(cnt) > 0 && zsign(sieve) != 0)
	{
		if (zeven(sieve))
		{
			p[length++] = ztouint(*n) - ztouint(nOld);
			if (zscompare(*n, MaxUpperBound) > 0)
				return;
			zcopy(*n, &nOld);
		}

		zsadd(*n, 2, n);
		zsdec(cnt, 2);
		zrshift(sieve, 1, &sieve);
	}
} // eval


/*
 * This is the original sieve algorithm that uses incr_list but not
 * primediff.dat (since we want to create primediff.dat).
 * Sieb des Eratosthenes
 * siebt, beginnend bei n, count zahlen unter verwendung der globalen liste
 * inctab. Liefert das ergebnis des
 * siebens, nachdem alle faktoren <= upper ausgesiebt sind.
 * ausgabe ist eine integerzahl, deren binaerdarstellung das sieb
 * darstellt (listen waeren viel langsamer).
 * lsb der ausgabe entspricht erster ungerader zahl des inter-
 * valls [n, n+count-1], naechstes bit der folgenden ungeraden zahl usw. ist
 * das bit 0, ist die entsprechende zahl prim, sonst zusammengesetzt.
 * ab kann auch ungerade sein, n > sqrt (n+count).
*/
void sieve(verylong n, verylong count, verylong upper, verylong* r)
{
	assert(zislong(upper));

	static verylong	temp = 0;
	static verylong	temp1 = 0;
	static verylong	s = 0;
	ulong		s1;
	static verylong	loops = 0;
	static verylong	first = 0;
	static verylong	last = 0;
	int		t = 1;			// Index into incr_list[]
	ulong		d = 3;
	ulong		h = d << 1;

	zadd(n, count, &last);
	zssub(last, 1, &temp);
	zsqrt(temp, &s, &temp1);
	assert(zcompare(n, s) >= 0);

	if (zcompare(s, upper) > 0)
		s1 = ztouint(upper);
	else
		s1 = ztouint(s);

	do
	{
		// Formula: loops = (count + d) / h;
		zsadd(count, d, &temp);
		zsdiv(temp, h, &loops);		// loops can become 0
		// Formula: first = d + h * ((n + d - 1) / h);
		zsadd(n, d - 1, &temp);
		zsdiv(temp, h, &temp1);
		zsmul(temp1, h, &temp);
		zsadd(temp, d, &first);

		if (zcompare(first, last) < 0)
		{
			// Formula: index = (ulong) ((first - n) / 2)
			zsub(first, n, &temp);
			z2div(temp, &temp);
			ulong index = ztouint(temp);

			do
			{
				zsetbit(r, index);
				index += d;
				zsdec(loops, 1);
			} while (zsign(loops) >= 0);
		}

		if (t >= incr_len)
			t = 5;			// New cycle
		d += incr_list[t++];
		h = d << 1;
	} while (d <= s1);

} // sieve


int main()
{
	static verylong temp = 0;

	InitBase();		// Ignore return value

	if (primeDiff != NULL)
	{
		// This is not likely to happen on the first run :-)
		int i = 0;
		verylong n = 0;

		zsinc(n, 2);
		zsinc(n, primeDiff[i++]);
		ztwo(&temp);

		do
		{
			if (strong_test(n, temp) != NumProbPrime)
				cout << "Error in primediff.dat, position " << i << endl;
			zsinc(n, primeDiff[i++]);
		} while(zscompare(n, 80000) < 0);

		zfree(&n);
	}

	p = (ulong *) malloc((size_t) (MaxUpperBound / 10 * sizeof(ulong)));
	if (p == NULL)
	{
		cout << "Not enough memory" << endl;
		return 1;
	}

	p[length++] = 1;	// 2 -> 3
	p[length++] = 2;	// 3 -> 5
	p[length++] = 2;	// 5 -> 7
	p[length++] = 4;	// 7 -> 11

	zuintoz(11, &nOld);

	verylong n = 0;
	verylong a = 0;
	verylong b = 0;
	zuintoz(100, &n);
	zuintoz(13, &b);
	sieve(b, n, b, &a);
	eval(a, &b, n);

	zuintoz(9000, &n);
	sieve(b, n, b, &a);
	eval(a, &b, n);

	zuintoz(20000, &n);
	while (zscompare(b, MaxUpperBound) < 0)
	{
		sieve(b, n, b, &a);
		eval(a, &b, n);
	}

	int fd = open("primediff.dat", O_CREAT | O_WRONLY, 0644);
	write(fd, p, length * sizeof(ulong));
	close(fd);
	free(p);

	return 0;
}

