#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "base.h"

// Searches for Carmichael numbers

const int MaxFactors = 64;

void root(verylong p, verylong start, ulong n, verylong* r)
{
	static verylong	old = 0;
	static verylong	temp = 0;
	static verylong	temp1 = 0;
	static verylong	temp2 = 0;

	zcopy(start, r);

	if (n == 2)
	{
		do
		{
			zcopy(*r, &old);
			zsq(old, &temp);
			zsub(temp, p, &temp);	// temp = old * old - p
			zsmul(old, n, &temp1);
			zdiv(temp, temp1, &temp, &temp1);
			// Formula: r = old - (old * old - p) / (old * n);
			zsub(old, temp, r);
			//cout << p << "**(1/2";
			//cout << ") = " << *r << endl;
		} while (zcompare(*r, old) != 0);
	}

	do
	{
		zcopy(*r, &old);
		zsexp(old, n - 1, &temp);
		zmul(old, temp, &temp2);
		zsub(temp2, p, &temp2);	// temp2 = old * temp - p
		zsmul(temp, n, &temp1);
		zdiv(temp2, temp1, &temp, &temp1);
		// Formula: r = old - (old * temp - p) / (temp * n);
		zsub(old, temp, r);
		//cout << p << "**(1/" << n;
		//cout << ") = " << *r << endl;
	} while (zcompare(*r, old) != 0);

}

void print(verylong p, verylong* d, ulong n)
{
	cout << p << " =";
	for (ulong j = 0; j < n; j++)
		cout << " " << d[j];
	cout << endl;
}

void NProd(ulong b, ulong n, int t, verylong* r)
{
	zone(r);

	if (trace_level > 5)
		cout << "NProd(" << b << ", " << n << ") = ";

	while (--n > 0)
	{
		b += primeDiff[t++];
		zsmul(*r, b, r);
	}

	if (trace_level > 5)
		cout << *r << endl;

}

int Usage(char** argv, int errCode)
{
	cout << "Usage: " << argv[0] << "start end" << endl;
	cout << "  start >= 0" << endl;
	cout << "  end >= start" << endl;
	return errCode;
}

int main(int argc, char** argv)
{
	verylong l = 0;
	verylong u = 0;
	verylong temp = 0;
	verylong temp1 = 0;

	if (InitBase() == false)
		return 1;

	if (argc < 3)
		return Usage(argv, 1);

	if (zsread(argv[1], &l) == 0 || zsign(l) < 0)
		return Usage(argv, 1);

	if (zsread(argv[2], &u) == 0 || zsign(u) == 0 || zcompare(u, l) <= 0)
		return Usage(argv, 1);

	trace_level = 5;

	//InitCheckMark(argv[0]);

	verylong d[MaxFactors];
	verylong lastD = 0;
	verylong limit[MaxFactors];
	verylong lastLimit = 0;
	verylong p[MaxFactors];
	verylong lastP = 0;
	verylong lastPlessOne = 0;
	verylong prevProd = 0;
	int t[MaxFactors];
	int lastT;
	verylong lowerbound = 0;
	ulong	n = 3;
	int i = 0;
	int j;


	// One time initialization.
	for (j = 0; j < MaxFactors; j++)
	{
		d[j] = 0;
		limit[j] = 0;
		p[j] = 0;
	}

	do	// For each number of factors 3,4,...
	{
		cout << "Searching for numbers with " << n << " factors" << endl;

		zuintoz(3, &d[0]);
		zuintoz(3, &p[0]);
		t[0] = 1;
		// Formula: limit[0] = u / NProd(d[0], n, t[0])
		NProd(ztouint(d[0]), n, t[0], &temp);
		zdiv(u, temp, &limit[0], &temp1);

		if (zscompare(limit[0], 3) < 0)
			break;

		do
		{
			while (++i < n - 1)
			{
				t[i] = t[i - 1];
				zsadd(d[i - 1], primeDiff[t[i]++], &d[i]);
				zmul(p[i - 1], d[i], &p[i]);
				NProd(ztouint(d[i]), n - i, t[i], &temp);
				zmul(p[i - 1], temp, &temp1);
				zdiv(u, temp1, &limit[i], &temp);
				zdiv(u, p[i - 1], &temp, &temp1);	// temp = u / p[i - 1]
				zsexp(limit[i], n - i, &temp1);
				if (zcompare(temp1, temp) > 0)
				{
					// Prop. 1.1
					root(temp, limit[i], n - i, &limit[i]);
				}
	#if 0
				// Checking Prop. 3.1 doesn't seem worth the effort.
				if (i == n - 2 && limit[i] > 2 * p[i - 1] * p[i - 1])
				{
					cout << "limit reduced by prop. 3.1" << endl;
					limit[i] = 2 * p[i - 1] * p[i - 1];
				}
	#endif
				assert(zscompare(limit[i], MaxUpperBound) <= 0);
			}

			zcopy(p[i - 1], &prevProd);
			zdiv(u, prevProd, &lastLimit, &temp);
			if (zcompare(lastLimit, prevProd) > 0)	// Prop. 1.3
				zsqrt(u, &lastLimit, &temp);

			zsexp(p[i - 2], 3, &temp);
			if (zcompare(lastLimit, temp) > 0)
				zcopy(temp, &lastLimit);	// Prop. 3.2

			assert(zscompare(lastLimit, MaxUpperBound) <= 0);
			lastT = t[i - 1];
			zcopy(d[i - 1], &lastD);
			// Formula: lowerbound = (l + prevProd - 1) / prevProd;
			zadd(l, prevProd, &temp);
			zsdec(temp, 1);
			zdiv(temp, prevProd, &lowerbound, &temp1);
			if (trace_level > 5)
				cout << "lowerbound = " << lowerbound << endl;
			do
			{
				zsinc(lastD, primeDiff[lastT++]);
			} while (zcompare(lastD, lowerbound) < 0);

			//cout << "lastD = " << lastD << ", lastLimit = " << lastLimit << endl;
			while (zcompare(lastD, lastLimit) <= 0)
			{
				zmul(prevProd, lastD, &lastP);
				zssub(lastP, 1, &lastPlessOne);
				zssub(lastD, 1, &temp);
				zdiv(lastPlessOne, temp, &temp, &temp1);
				if (zsign(temp1) == 0)
				{
					for (j = 0; j < i; j++)
					{
						zssub(d[j], 1, &temp);
						zdiv(lastPlessOne, temp, &temp, &temp1);
						if (zsign(temp1) != 0)
							break;
					}

					if (j == i)
					{
						zcopy(lastD, &d[i]);
						print(lastP, d, n);
					}
				}
				zsinc(lastD, primeDiff[lastT++]);
			}

			do
			{
				i--;
				zsinc(d[i], primeDiff[t[i]++]);
				if (i > 0)
					zmul(d[i], p[i - 1], &p[i]);
				else
					zcopy(d[i], &p[i]);
			} while (i > 0 && zcompare(d[i], limit[i]) > 0);
		} while (zcompare(d[i], limit[i]) <= 0);

		n++;
	} while (1);

	return 0;

}

