#include <assert.h>
#include "base.h"





IntList::IntList()
{
	mHead = NULL;
	mTail = NULL;
	mSize = 0;
} // IntList





IntList::~IntList()
{
	Clear();
} // ~IntList





void IntList::Clear()
{
	while (mSize != 0)
	{
		Remove();
	}
} // Clear





void IntList::Remove()
{
	assert(mHead != NULL);
	assert(mTail != NULL);
	assert(mSize != 0);

	IntElem* elem = mTail;
	mTail = elem->mPrev;

	mSize--;
	assert(mSize >= 0);
	if (mSize != 0)
	{
		assert(mHead != NULL);
		assert(mTail != NULL);
		mTail->mNext = NULL;
	}
	else
	{
		assert(mTail == NULL);
		mHead = NULL;
	}

	delete elem;

} // Remove





IntElem* IntList::GetNext(POSITION& pos) const
{
	assert(pos != NULL);
	assert(pos->mPrev == NULL || pos->mPrev->mNext == pos);
	assert(pos->mNext == NULL || pos->mNext->mPrev == pos);

	IntElem* elem = pos;
	pos = pos->mNext;

	return elem;

} // GetNext





bool IntList::Add(verylong n, ushort count, ushort state)
{
	return Add(new IntElem(n, count, state));
} // Add




bool IntList::Add(ulong n, ushort count, ushort state)
{
	return Add(new IntElem(n, count, state));
} // Add





bool IntList::Add(IntElem* elem)
{
	assert(elem != NULL);

	if (mTail == NULL)
	{
		assert(mHead == NULL);
		assert(mSize == 0);
		mHead = elem;
	}
	else
	{
		assert(mHead != NULL);
		assert(mSize != 0);
		assert(mTail->mNext == NULL);
		mTail->mNext = elem;
	}

	elem->mPrev = mTail;	// May be NULL
	elem->mNext = NULL;
	mTail = elem;
	mSize++;
	return true;

} // Add





ostream& operator << (ostream& o, const IntList& l)
{
	POSITION pos = l.GetHeadPosition();

	while (pos != NULL)
	{
		IntElem* p = l.GetNext(pos);
		int j = 0;
		while (j++ < p->GetCount())
			o << " " << *p->GetNum();
	}

	return o;

} // <<





void IntList::Multiply(verylong* r) const
{
	POSITION	pos = GetHeadPosition();
	static verylong	temp = 0;

	zone(r);

	while (pos != NULL)
	{
		IntElem* p = GetNext(pos);
		zsexp(*p->GetNum(), p->GetCount(), &temp);
		zmulin(temp, r);
	}

} // Multiply





static bool notprime(verylong n)
{
	// Do 100 pseudo prime tests.
	if (zmcomposite(n, 100))
		return true;

	return false;

} // notprime





bool IntList::Verify(verylong n) const
{
	static verylong temp = 0;
	static verylong quot = 0;
	static verylong rem = 0;
	static verylong prod = 0;

	if (trace_level > 0)
		cout << "Verifying factor list...";

	zone(&prod);

	POSITION pos = GetHeadPosition();
	while (pos != NULL)
	{
		IntElem* p = GetNext(pos);
		if (p->GetState() != NumPrime
			|| p->GetCount() == 0
			|| p->GetCount() > 10000)
		{
			cout << "The factorlist is corrupt" << endl;
			return false;
		}

		zsexp(*p->GetNum(), p->GetCount(), &temp);
		zmulin(temp, &prod);
		zdiv(n, temp, &quot, &rem);
		if (zsign(rem) != 0)
		{
			cout << "The factorlist contains a non-factor ";
			cout << *p->GetNum() << endl << " of ";
			cout << n << endl;
			return false;
		}

		if (notprime(*p->GetNum()))
		{
			cout << "The factor " << *p->GetNum();
			cout << " is not prime" << endl;
			return false;
		}
	} // while

	if (zcompare(n, prod) != 0)
	{
		cout << "The product of the factors is " << prod;
		cout << " which doesn't equal " << n << endl;
		return false;
	}

	if (trace_level > 0)
		cout << " done" << endl;

	return true;

} // Verify





void IntList::Merge(IntList& l1, IntList& l2)
{
	// Sort l1, l2 into this list.
	IntElem* pos1 = l1.mHead;
	IntElem* pos2 = l2.mHead;
	IntElem* temp;

	while (pos1 != NULL && pos2 != NULL)
	{
		if (zcompare(*pos1->mNum, *pos2->mNum) < 0)
		{
			temp = pos1;
			pos1 = pos1->mNext;
			Add(temp);
		}
		else
		{
			temp = pos2;
			pos2 = pos2->mNext;
			Add(temp);
		}
	}

	while (pos1 != NULL)
	{
		temp = pos1;
		pos1 = pos1->mNext;
		Add(temp);
	}

	while (pos2 != NULL)
	{
		temp = pos2;
		pos2 = pos2->mNext;
		Add(temp);
	}

	l1.mHead = NULL;
	l1.mTail = NULL;
	l1.mSize = 0;
	l2.mHead = NULL;
	l2.mTail = NULL;
	l2.mSize = 0;

} // Merge
