#include <stdio.h>
#include <stdlib.h>
#include "base.h"

// Test primality of Mersenne numbers

int Usage(char** argv, int errorCode)
{
	cout << "Usage: " << argv[0] << " exponent" << endl;
	cout << "  exponent > 2" << endl;
	return errorCode;
}

int main(int argc, char** argv)
{
	ulong n;

	trace_level = 0;
	if (InitBase() == false)
		return 1;

	if (argc < 2)
		return Usage(argv, 1);

	n = atol(argv[1]);
	if ((long) n < 3)
		return Usage(argv, 1);

	verylong m = 0;
	verylong two = 0;

	ztwo(&two);
	zsetbit(&m, n);
	zsdec(m, 1);		// m = 2**n - 1 = M(n)

	if (strong_test(m, two) == NumComposite)
	{
		cout << "M(" << n << ") is composite according to strong Fermat test" << endl;
		return 0;
	}

	verylong s = 0;
	zuintoz(4, &s);
	ulong k = n - 1;

	zmstart(m);

	ztom(s, &s);
	ztom(two, &two);

	while (--k > 0)
	{
		zmontsq(s, &s);
		zmontsub(s, two, &s);
	}

	zmtoz(s, &s);
	zmfree();

	if (zsign(s) == 0)
		cout << "M(" << n << ") is prime" << endl;
	else
		cout << "M(" << n << ") is composite" << endl;

	return 0;

}

