#include <assert.h>
#include "base.h"

const ulong DIFNPMAX = 100;


/* Performs the standard 2nd phase */
/* p is the least prime greater than b1 */
bool pplus1step2(verylong n, verylong a, ulong p, int t,
	ulong b2, verylong* factor)
{
	static verylong	prod = 0;
	static verylong	a1 = 0;
	static verylong	temp = 0;
	static verylong	temp1 = 0;
	int		nbp = 0;
	ulong		nbgcd = b2 / 100;
	ulong		pDiff;

	if (trace_level > 1)
		cout << "pplus1step2: n = " << n
			<< ", b2 = " << b2 << endl;

	if (p <= 3)
		return false;

	if (nbgcd == 0)
		nbgcd = 1;

	zcopy(a, &a1);
	zcopy(a1, &temp1);
	lucas2(p, n, temp1, &temp, &a1);
	zssub(a1, 2, &prod);

	do
	{
		pDiff = primeDiff[t - 1] / 2;
		assert(pDiff < DIFNPMAX);
		zcopy(a1, &temp1);
		lucas2(pDiff, n, temp1, &temp, &a1);
		if (zsign(a1) == 0)
		{
			// I don't know if this is a bug or normal behaviour.
			cout << "WARNING: a1 is 0 in pplus1step2" << endl;
		}
		// Formula: prod = (prod * (a1 - 2)) % n;
		zssub(a1, 2, &temp);
		zmulmod(temp, prod, n, &prod);

		if (++nbp % nbgcd == 0)
		{
			zgcd(prod, n, factor);
			if (zscompare(*factor, 1) > 0 && zcompare(*factor, n) < 0)
				return true;
		}

		p += primeDiff[t++];
	} while (p <= b2);

	if (nbp % nbgcd)
	{
		if (trace_level > 1)
			cout << "  Performing last gcd in step 2" << endl;
		zgcd(prod, n, factor);
		if (zscompare(*factor, 1) > 0 && zcompare(*factor, n) < 0)
			return true;
	}

	return false;

} // pplus1step2





ushort pplus1_factor(verylong n,
	ulong b1, ulong b2, verylong* factor)
{
	// The values of these are no longer needed after recursion.
	static verylong	a = 0;
	static verylong	a1 = 0;

	static verylong	temp = 0;
	bool		done = false;
	int		t = 1;
	ulong		nbp = 0;
	ulong		p = 3;		// Current divisor
	ulong		q;
	ulong		m;
	ulong		nbgcd = b1 / 100;

	if (trace_level > 0)
		cout << "P+1: N = " << n
			<< ", b1 = " << b1 << ", b2 = " << b2 << endl;

	if (nbgcd == 0)
		nbgcd = 1;

	zuintoz(3, &a);		// 2 does not work

	// First handle base 2.
	q = 2;
	m = b1 / q;

	while (q < m)
		q <<= 1;

	zcopy(a, &a1);
	lucas2(q, n, a1, &temp, &a);

	do
	{
		nbp++;
		m = b1 / p;
		q = p;
		while (q < m)
			q *= p;
		zcopy(a, &a1);
		lucas2(q, n, a1, &temp, &a);
		if (nbp % nbgcd == 0)
		{
			zssub(a, 2, &temp);
			zgcd(temp, n, factor);
			if (zscompare(*factor, 1) > 0)
			{
				if (trace_level > 1)
					cout << "  Step 1 found factor " << *factor
						<< " at p = " << p << endl;
				if (p > 3 && zcompare(*factor, n) == 0)
				{
					p = 63 * p / 64;
					pplus1_factor(n, p, b2, factor);
					// The resulting factor can be 1. Shit happens.
				}
				done = true;
				nbp = nbgcd;	// Prevent duplicate test
				break;
			}
		}
		p += primeDiff[t++];
	} while (p <= b1);

	if (nbp % nbgcd)
	{
		if (trace_level > 1)
			cout << "  Performing last gcd in step 1" << endl;
		zssub(a, 2, &temp);
		zgcd(temp, n, factor);
		if (zscompare(*factor, 1) > 0)
		{
			if (trace_level > 1)
				cout << "  Step 1 found factor " << *factor
					<< " at p = " << p << endl;
			if (zcompare(*factor, n) == 0)
			{
				pplus1_factor(n, b1 / 2, b2, factor);
				// The resulting factor can be 1. Shit happens.
			}
			done = true;
		}
	}

	while (!done)
	{
		done = pplus1step2(n, a, p, t, b2, factor);
		if (done && trace_level > 4)
			cout << "pplus1step2 found factor " << *factor << endl;

		if (done && zcompare(*factor, n) == 0)
		{
			// It would be more efficient to loop within step2.
			done = false;
			b2 /= 2;
			if (b2 < 10)
				break;		// Not successful
		}
		else
		{
			break;
		}
	}

	if (done && zscompare(*factor, 1) > 0 && zcompare(*factor, n) < 0)
		return NumPrime;

	zone(factor);

	return NumComposite;

} // pplus1_factor
